# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

from mml_tags.tags.confuse import confuse_task
from mml_tags.tags.redistribute_splits import redistribute_splits
from mml_tags.tags.shrink_train import shrink_train
from mml_tags.tags.subclass import subclass_task
from mml_tags.tags.subset import subset_task

from mml.core.data_preparation.task_creator import TASK_CREATOR_TAG_MAP, TaskCreator

TaskCreator.shrink_train = shrink_train
TaskCreator.confuse_task = confuse_task
TaskCreator.redistribute_splits = redistribute_splits
TaskCreator.subclass_task = subclass_task
TaskCreator.subset_task = subset_task

TASK_CREATOR_TAG_MAP["shrink_train"] = "shrink_train"
TASK_CREATOR_TAG_MAP["redistribute_splits"] = "redistribute_splits"
TASK_CREATOR_TAG_MAP["confuse"] = "confuse_task"
TASK_CREATOR_TAG_MAP["subset"] = "subset_task"
TASK_CREATOR_TAG_MAP["subclass"] = "subclass_task"



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

import logging
from collections import Counter

import numpy as np

from mml.core.data_loading.task_attributes import Modality, TaskType
from mml.core.data_preparation.task_creator import TaskCreator, implements_action
from mml.core.data_preparation.utils import TaskCreatorActions

logger = logging.getLogger(__name__)


@implements_action(TaskCreatorActions.SET_FOLDING)
def subset_task(self: TaskCreator, val_string: str) -> None:
    """
    Simple version of subsetting a task. Every fold will be deterministically shrunk to roughly val_string
    * 100 percent of samples. This process will not take class balance into consideration.

    :param val_string: string of a float within (0, 1), replace comma by underscore (e.g. 0_01 for 1%)
    :return: False, since by sub-setting heavy stats shifts may occur
    """
    percent = float(val_string.replace("_", ".")) * 100
    assert 0 < percent < 100, f"was given tag subset with {val_string=}, must be in range (0, 1)"
    logger.info(f"Subsetting task {self.current_meta.name} to {percent}% of training data (ensuring folds).")
    self.protocol(f"Subset with value {val_string}")
    # set random seed to ensure reproducibility
    np.random.seed(42)
    for fold_ix, fold in enumerate(self.current_meta.train_folds):
        # version 1
        size = int(percent * len(fold) / 100)
        if size == 0:
            msg = f"Subsetting to {percent}% would result in 0 samples within fold {fold_ix}"
            logger.critical(msg)
            raise RuntimeError(msg)
        new_fold = np.random.choice(a=fold, size=size, replace=False).tolist()
        assert abs((len(new_fold) * 100 / len(fold)) - percent) <= 1, "subsetting was not successful"
        logger.debug(
            f"Shrunk fold {fold_ix} from {len(fold)} to {len(new_fold)} equalling roughly "
            f"{len(new_fold) * 100 / len(fold):.2f}%."
        )
        # Remove unused samples
        unused = [sample_id for sample_id in self.current_meta.train_folds[fold_ix] if sample_id not in set(new_fold)]
        for sample_id in unused:
            self.current_meta.train_samples.pop(sample_id)
        # finally set the fold
        self.current_meta.train_folds[fold_ix] = new_fold
    # re-compute class occurrences
    if self.current_meta.task_type == TaskType.CLASSIFICATION:
        self.current_meta.class_occ = Counter(
            [
                self.current_meta.idx_to_class[sample[Modality.CLASS]]
                for sample in self.current_meta.train_samples.values()
            ]
        )
        # we might have lost some classes due to subsetting
        for cls in self.current_meta.idx_to_class.values():
            if cls not in self.current_meta.class_occ:
                self.current_meta.class_occ[cls] = 0



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

from typing import List

from mml.core.data_preparation.task_creator import TaskCreator, implements_action
from mml.core.data_preparation.utils import TaskCreatorActions


@implements_action(TaskCreatorActions.NONE)
def subclass_task(self: TaskCreator, *classes_to_keep: List[str]) -> None:
    """
    Restricts the task to the classes given. This may be either ...
    :param classes_to_keep:
    :return:
    """
    # TODO implement functionality
    # requires remapping in idx_to_class!
    # requires new class_occ
    raise NotImplementedError()



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

import logging

import numpy as np
import numpy.random

from mml.core.data_loading.task_attributes import Modality, TaskType
from mml.core.data_preparation.task_creator import TaskCreator, implements_action
from mml.core.data_preparation.utils import TaskCreatorActions

logger = logging.getLogger(__name__)

MIN_OCC_PER_CLASS = 5  # constant representing the minimum occ per class


@implements_action(TaskCreatorActions.SET_STATS)
def shrink_train(self: TaskCreator, val_string: str, seed: str = "42") -> None:
    """
    Alternative version of sub-setting a task. Every fold but fold 0 (which will be used for evaluation)
    is deterministically shrunk to sum all remaining training samples to total val_string.

    MIN_OCC_PER_CLASS ensures (if sufficient samples of a class are available) a minimum number of samples of a
    class.

    :param val_string: string of an int within [(num_folds - 1)*5*num_classes, (num_folds - 1) * num_samples / num_folds]
    :return: None, influences current_meta attribute
    """
    if self.current_meta.task_type != TaskType.CLASSIFICATION:
        raise NotImplementedError(f"shrink_train is not implemented for task type {self.current_meta.task_type}")
    new_total_class_occ = {val: 0 for val in set(self.current_meta.idx_to_class.values())}
    num_folds = len(self.current_meta.train_folds)
    max_samples = sum([len(fold) for fold in self.current_meta.train_folds[1:]])
    min_samples = (num_folds - 1) * MIN_OCC_PER_CLASS * len(new_total_class_occ)
    remaining = int(val_string)
    assert (
        min_samples <= remaining <= max_samples
    ), f"was given tag subset with {val_string=}, must be in range [{min_samples}, {max_samples}]"
    logger.info(
        f"Shrinking training folds of task {self.current_meta.name} to {remaining} of training "
        f"data (ensuring folds). Seed is: {seed}."
    )
    self.protocol(f"Shrink training with value {val_string} and seed {seed}")
    for fold_ix, fold in enumerate(self.current_meta.train_folds):
        # skip first fold, this will be the validation / test split
        if fold_ix == 0:
            continue
        # set random seed to ensure reproducibility
        rng = numpy.random.default_rng(int(seed) + fold_ix * 42)
        # calculate remaining size for this fold
        size = remaining // (num_folds - 1)
        if remaining % (num_folds - 1) > fold_ix:
            size += 1
        assert 1 <= size <= len(fold), f"Size requirement for fold {fold_ix} not met, {size=} but {len(fold)=}."
        # assert all classes have enough occurrences
        pre_selection = set()
        for cls in new_total_class_occ:
            cls_samples = [
                sample_id
                for sample_id in fold
                if self.current_meta.idx_to_class[self.current_meta.train_samples[sample_id][Modality.CLASS.value]]
                == cls
            ]
            if len(cls_samples) < MIN_OCC_PER_CLASS:
                logger.warning(f"Only {len(cls_samples)} samples for class {cls} " f"(requires {MIN_OCC_PER_CLASS}).")
                pre_selection.update(cls_samples)
            else:
                pre_selection.update(rng.choice(a=cls_samples, size=MIN_OCC_PER_CLASS, replace=False).tolist())
        # now fill the remaining gap
        if size - len(pre_selection) <= 0:
            raise RuntimeError("Too many classes to ensure sufficient samples per class in each fold.")
        non_selected = [sample_id for sample_id in fold if sample_id not in pre_selection]
        post_selection = rng.choice(a=non_selected, size=size - len(pre_selection), replace=False).tolist()
        new_fold = np.random.permutation(list(pre_selection) + post_selection).tolist()
        new_fold_class_occ = {val: 0 for val in set(self.current_meta.idx_to_class.values())}
        for sample_id in new_fold:
            new_fold_class_occ[
                self.current_meta.idx_to_class[self.current_meta.train_samples[sample_id][Modality.CLASS]]
            ] += 1
        if not all([occ >= MIN_OCC_PER_CLASS for occ in new_fold_class_occ.values()]):
            logger.warning(
                f"not sufficient samples to guarantee {MIN_OCC_PER_CLASS} occ per class in "
                f"fold {fold_ix}, resulting occ was {new_fold_class_occ}."
            )
        else:
            # successful shrinkage
            logger.debug(
                f"was able to guarantee {MIN_OCC_PER_CLASS} occ per class in fold {fold_ix}, "
                f"resulting occ was {new_fold_class_occ}."
            )
        logger.debug(
            f"Shrunk fold {fold_ix} from {len(fold)} to {len(new_fold)} equalling roughly "
            f"{len(new_fold) * 100 / len(fold):.2f}%."
        )
        # Remove unused samples
        unused = [sample_id for sample_id in self.current_meta.train_folds[fold_ix] if sample_id not in set(new_fold)]
        for sample_id in unused:
            self.current_meta.train_samples.pop(sample_id)
        # update total class_occ
        for k, v in new_fold_class_occ.items():
            new_total_class_occ[k] += v
        # finally set the fold
        self.current_meta.train_folds[fold_ix] = new_fold
    # set new total class occ, this one ignores the validation split!
    self.current_meta.class_occ = new_total_class_occ



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

VERSION = (0, 6, 1)
__version__ = ".".join(map(str, VERSION))
__all__ = ["__version__"]



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

from mml_tags import __version__

from mml.core.data_preparation.task_creator import TASK_CREATOR_TAG_MAP, TaskCreator
from mml.core.scripts.utils import ARG_SEP, TAG_SEP, load_mml_plugins


def list_available_tags():
    load_mml_plugins()
    print(f"mml_tags plugin, version {__version__}")
    print(f"Available tags (total {len(TASK_CREATOR_TAG_MAP)}):")
    for tag_id, tag_fct in TASK_CREATOR_TAG_MAP.items():
        print(f"  * {tag_id}\n{getattr(TaskCreator, tag_fct).__doc__}\n")
    print(
        f"\nApply tags with with tag seperator {TAG_SEP} and argument seperator {ARG_SEP}. See configs/tasks for "
        f"more details."
    )



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

import logging
from collections import Counter

from mml.core.data_loading.task_attributes import DataSplit, Modality, TaskType
from mml.core.data_preparation.task_creator import TaskCreator, implements_action
from mml.core.data_preparation.utils import TaskCreatorActions

logger = logging.getLogger(__name__)


@implements_action(TaskCreatorActions.SET_STATS)
def redistribute_splits(self: TaskCreator, *new_distribution: str) -> None:
    """
    Allows for redistribution of the train-validation-test splits of a task.

    Receives either 1, 2 or 3 arguments that are interpreted as train(:val) or train:val:test percentages.
    Values must be floats (given as e.g. '0_25'), in case of 3 arguments they must sum 1, otherwise raises a
    :exc:ValueError.

    If one value is provided the test split is left untouched, all train tuples are redistributed into folds,
    whereas conserving the number of folds, but fold 1 ... fold n receive together the fraction of samples given
    in the argument while fold 0 (which is interpreted as validation fold) receives the rest. Be aware that for
    classification tasks balanced classes are tried to be ensured.

    If two values are provided the original test split is dropped, if the two values do not sum up to one, the
    remaining fraction is (randomly) selected as new test set (from the original training split). For the rest the
    method behaves as in the case with one argument - first value fraction will be distributed across folds 1, ...,
    n and the second value fraction become fold 0.

    For the variant with three arguments ALL original samples (train and test) are put into a pool and test split
    is resampled from this pool (based on the third value fraction). Afterward with the remaining samples the
    procedure is identical to the other argument variants, except as the folds are built from the remaining pool.

    :param new_distribution: either one, two or three strings that represent floats for distributing samples
    :return: None
    """
    # log
    logger.info(f"Redistributing data of task {self.current_meta.name} with {new_distribution=}.")
    self.protocol(f"Redistributed with value(s) {new_distribution}.")
    # check args
    if len(new_distribution) < 1 or len(new_distribution) > 3:
        raise ValueError("Redistribute tag must be provided with either 1, 2 or 3 arguments (split by ' ')")
    new_distribution = [float(elem.replace("_", ".")) for elem in new_distribution]
    drop_original_test = len(new_distribution) == 2
    if drop_original_test:
        test_frac = 1 - sum(new_distribution)
        if test_frac < 0 or test_frac >= 1:
            raise ValueError("Redistribute tag with two arguments must sum within (0, 1].")
        new_distribution.append(test_frac)
    modify_test = len(new_distribution) == 3
    if modify_test and sum(new_distribution) != 1:
        raise ValueError("Redistribute tag with three arguments must sum up to one.")
    if not modify_test and (sum(new_distribution) <= 0 or sum(new_distribution) >= 1):
        raise ValueError("Redistribute tag with one argument must be within (0, 1) excluding borders.")
    if modify_test and self.current_meta.task_type != TaskType.CLASSIFICATION:
        raise RuntimeError(
            "Since class occurrences are changed if train/test set changes, this is currently "
            "only available for classification tasks."
        )
    if modify_test and self.fm.preprocessed_data in self.dset_path.parents:
        raise RuntimeError(
            "Since redistribute tag modifies the test data, this can only be run on a raw (none "
            "preprocessed) version of a task. If this task already has been preprocessed with "
            "some preprocessing id X, consider creating the task with some dummy preprocessing id "
            "e.g. mml info (task settings as before) preprocessing=example and run your "
            "original command afterwards."
        )
    # set self.data['train'] and self.data['test'], recalculate class_occ if necessary
    old_folds_n = len(self.current_meta.train_folds)
    if modify_test:
        all_ids = list(self.current_meta.train_samples.keys())
        all_samples = self.current_meta.train_samples
        if not drop_original_test:
            all_ids.extend(list(self.current_meta.test_samples.keys()))
            all_samples.update(self.current_meta.test_samples)
        # use split folds mechanism to extract the new test split balanced
        self.data = {DataSplit.FULL_TRAIN: all_samples}
        self.split_folds(n_folds=2, ensure_balancing=True, fold_0_fraction=new_distribution[2], ignore_state=True)
        train_ids = self.current_meta.train_folds[1]
        test_ids = self.current_meta.train_folds[0]
        self.data = {
            DataSplit.FULL_TRAIN: {s_id: all_samples[s_id] for s_id in train_ids},
            DataSplit.TEST: {s_id: all_samples[s_id] for s_id in test_ids},
        }
        # update class occurrences
        self.current_meta.class_occ = Counter(
            [self.current_meta.idx_to_class[all_samples[s_id][Modality.CLASS]] for s_id in train_ids]
        )
    else:
        self.data = {
            DataSplit.FULL_TRAIN: self.current_meta.train_samples,
            DataSplit.TEST: self.current_meta.test_samplestest_tuples,
        }
    # calculate validation split and run self.split_folds
    if modify_test:
        fraction = new_distribution[1] / (new_distribution[0] + new_distribution[1])
    else:
        fraction = 1 - new_distribution[0]
    self.split_folds(n_folds=old_folds_n, ensure_balancing=True, fold_0_fraction=fraction, ignore_state=True)



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

import logging
from itertools import chain

import numpy as np

from mml.core.data_loading.task_attributes import Modality, TaskType
from mml.core.data_preparation.task_creator import TaskCreator, implements_action
from mml.core.data_preparation.utils import TaskCreatorActions

logger = logging.getLogger(__name__)


@implements_action(TaskCreatorActions.NONE)
def confuse_task(self: TaskCreator, val_string: str) -> None:
    """
    Simple version of making a task confusing by setting a fraction of labels as random.

    :param val_string: string of a float within (0, 1), replace comma by underscore (e.g. 0_01 for 1%)
    :return: None
    """
    if self.current_meta.task_type != TaskType.CLASSIFICATION:
        raise RuntimeError("confuse tag only valid for classification task")
    percent = float(val_string.replace("_", ".")) * 100
    assert 0 < percent < 100, f"was given tag confuse with {val_string}, must be in range (0, 1)"
    logger.info(f"Confusing {percent}% of training data of task {self.current_meta.name}.")
    self.protocol(f"Confused with value {val_string}")
    # ensure reproducibility
    np.random.seed(42)
    all_ids = list(chain(*self.current_meta.train_folds))
    confuse_ids = np.random.choice(a=all_ids, size=int(percent * len(all_ids) / 100), replace=False).tolist()
    assert abs((len(confuse_ids) * 100 / len(all_ids)) - percent) <= 1, "confusing was not successful"
    num_classes = len(set(self.current_meta.idx_to_class.values()))
    for confuse_id in confuse_ids:
        new_val = np.random.randint(num_classes)
        old_val = self.current_meta.train_samples[confuse_id][Modality.CLASS.value]
        self.current_meta.train_samples[confuse_id][Modality.CLASS.value] = new_val
        self.current_meta.class_occ[self.current_meta.idx_to_class[old_val]] -= 1
        self.current_meta.class_occ[self.current_meta.idx_to_class[new_val]] += 1
    logger.debug(
        f"Confused {len(confuse_ids)} out of {len(all_ids)} training_tuples equalling roughly "
        f"{len(confuse_ids) * 100 / len(all_ids)}%."
    )
